"""Materialien zu den zentralen NRW-Abiturprüfungen im Fach Informatik ab 2018
implementiert in Python.
"""

from __future__ import annotations

__version__: Final[str] = "1.4.0"

from typing import Final
